import json
import chattie.inventory as inventory
from os.path import isfile


class Inventory(inventory.Inventory):
    """An inventory a class that has a get and set method serving
    as a key value store.

    This inventory is a simple JSON file dict-backed implementation and is
    simply a reference implementation for making more complex and useful
    inventories.
    """
    inventory = {}

    def __init__(self):
        """Init initializes the inventory."""
        if isfile("./inventory.json"):
            print("Loading my inventory from last time...")
            self.__load_inventory()

    def get(self, key):
        """Get the value for key."""
        return self.inventory.get(key, None)

    def set(self, key, value):
        """Set key to value."""
        self.inventory[key] = value
        self.__save_inventory()

    def __load_inventory(self):
        """Load the inventory file from the filesystem.

        Potentially destructive function so we attempt to privatize it.
        """
        with open("./inventory.json", "r") as inv:
            self.inventory = json.load(inv)

    def __save_inventory(self):
        """Save the inventory to the file system.

        Potentially destructive function so we attempt to privatize it.
        """
        with open("./inventory.json", "w") as inv:
            json.dump(self.inventory, inv)



"""The primary Bot class which handles inventory and connections."""

import os
import sys
from os.path import exists
from os.path import isfile


class Bot:
    """Base Bot class, maintains state and parsing commands."""

    def __init__(self, name, inventory, command_pkgs, handlers=[]):
        """Initialize the bot.

        command_pkgs should be a list of packages as returned by
        get_commands() from chattie.plugins. (essentially as returned
        by the entry_points functions)

        A command package needs to have a global dict variable named
        commands which contains a key for each command name and a
        corresponding value which is the function to call for that
        command. The command functions will be called with two
        arguments the first being the current instance of the Bot
        class the second will be an argv like array of the message.

        See the examples directory for commands, connectors, and handlers
        """
        print("Booting systems...")
        self.name = name
        print("Hello my name is " + name + "...")
        self.inventory = inventory
        print("Inventory gathered...")

        self.handlers = handlers
        self.commands = {}
        for pkg in command_pkgs:
            loaded = pkg.load()
            self.commands.update(loaded.commands)

        # Add current directory PYTHONPATH for dynamic imports.
        sys.path.append(os.getcwd())

        # Check if tricks exists and add it if so.
        if exists('./tricks') or isfile('./tricks.py'):
            import tricks
            self.commands.update(tricks.commands)

        # Look for local handlers
        if exists('./handlers') or isfile('./handlers.py'):
            import handlers
            self.handlers += handlers.handlers

        print("Commands and handlers loaded...")

    def parse_message(self, msg, extras={}):
        """Parse the message, run any commands and handlers found. Returns an
        array of strings to send back to the channel."""
        print("Message received...")
        messages = []

        if self.name.lower() in msg.lower():
            print("Someone is talking to me...")
            split = msg.split(" ")

            # Start the cmd_idx at 1
            cmd_idx = 1

            # get the first word after our name as that will be the
            # command always.
            for i, w in enumerate(split):
                if w.lower().endswith(self.name.lower()):
                    cmd_idx = i + 1
                    break

            cmd = self.commands.get(split[cmd_idx])
            if cmd is None:
                return ["I don't know that trick."]

            resp = cmd(self, split, **extras)
            if resp:
                return [resp]

            return []

        # If no command then pass to handlers
        for h in self.handlers:
            reply = h(self, msg)
            if reply:
                messages.append(reply)

        return messages

    def get(self, key):
        """Alias to bot.inventory.get"""
        print("DEPRECATION WARNING: bot.get is deprecated, please use"
              " bot.inventory.get instead.")
        return self.inventory.get(key)

    def set(self, key, value):
        """Alias to bot.inventory.set"""
        print("DEPRECATION WARNING: bot.set is deprecated, please use"
              " bot.inventory.set instead.")
        return self.inventory.set(key, value)



"""The reference Connector class that defines the required interface.

A Connector connects to any chat or backend service and handles delivering
incoming messages to the bot's parse_message method. It will then be
responsible for sending the responses from the bot back to the server."""

from abc import abstractmethod


class Connector:
    """The base interface that must be implemented by a backend."""

    @abstractmethod
    def __init__(self, bot):
        """Bot is the parse_message function of the Bot class.

        It should be passed the room_id (whatever form that takes) and
        the plain text of the incoming message from the service.
        """
        self.bot = bot

    @abstractmethod
    def listen(self):
        """Should connect and listen to incoming messages from the backend.

        When an incoming message is parsed should send to self.parser
        (the Bot class' parse_message method)
        """
        raise NotImplementedError



"""A Chattie connector for Matrix.

This connector requires that you register an account for the bot
before setting up a connection

Requires the following environment variables:

MATRIX_URL -- The Matrix homeserver, such as https://matrix.org
MATRIX_USERNAME -- The bot username
MATRIX_PASSWORD -- The bot password
MATRIX_ROOMS -- a comma seperated list of rooms for the bot to join
"""

import os
import chattie.connector as connector

try:
    from matrix_client.client import MatrixRequestError
    from matrix_client.client import MatrixClient
except ImportError:
    import sys
    print('You need to pip3 install matrix_client before '
          'using this connector!')
    sys.exit(1)

MATRIX_URL = os.getenv("MATRIX_URL")
MATRIX_USERNAME = os.getenv("MATRIX_USERNAME")
MATRIX_PASSWORD = os.getenv("MATRIX_PASSWORD")
MATRIX_ROOMS = os.getenv("MATRIX_ROOMS")


class Connector(connector.Connector):
    """A matrix connector for the Chattie bot framework."""

    def __init__(self, bot):
        """Connect to the matrix room."""
        self.bot = bot

        self.client = MatrixClient(MATRIX_URL)
        # Try to register if it fails try to log in.
        try:
            self.token = self.client.\
                         register_with_password(username=MATRIX_USERNAME,
                                                password=MATRIX_PASSWORD)
        except MatrixRequestError:
            self.token = self.client.\
                         login_with_password(username=MATRIX_USERNAME,
                                             password=MATRIX_PASSWORD)
        for room in MATRIX_ROOMS.split(","):
            room_name = room
            if not room.startswith("#"):
                room_name = "!" + room
            r = self.client.join_room(room_name)
            r.add_listener(self.parse_event)

    def listen(self):
        """Listen for messages on Matrix."""
        self.client.listen_forever()

    def parse_event(self, room, event):
        """Transform Matrix incoming_event to text."""
        if event['type'] != 'm.room.message':
            return

        resp = self.bot.parse_message(event['content']['body'])

        for r in resp:
            room.send_text(r)
